use std::collections::HashMap;
use std::sync::Arc;

use super::render::vulkan_context::VulkanContext;
use super::scene::SceneContext;
use crate::core::input::InputManager;
use crate::core::scene::SceneManager;
use crate::core::timer::Timer;
use crate::game::scenes::test_scene::TestScene;
use egui_winit_vulkano::{Gui, GuiConfig};
use user_event::UserEvent;
use vulkano::format::Format;
use vulkano::image::ImageUsage;
use vulkano::swapchain::PresentMode;
use vulkano_util::context::VulkanoContext;
use vulkano_util::window::{VulkanoWindows, WindowDescriptor};
use winit::application::ApplicationHandler;
use winit::event::WindowEvent;
use winit::event_loop::{ActiveEventLoop, EventLoopProxy};
use winit::window::WindowId;

pub mod user_event;

pub const DEPTH_IMAGE_ID: usize = 0;

pub struct App {
    vulkan_context: Arc<VulkanContext>,
    vulkano_windows: Arc<VulkanoWindows>,
    gui: HashMap<WindowId, Gui>,
    scene_manager: HashMap<WindowId, SceneManager>,
    input_manager: Arc<InputManager>,
    timer: Arc<Timer>,
    event_loop_proxy: EventLoopProxy<UserEvent>,
}

impl App {
    pub fn new(
        vulkano_context: VulkanoContext,
        input_manager: InputManager,
        event_loop_proxy: EventLoopProxy<UserEvent>,
    ) -> Self {
        Self {
            vulkan_context: Arc::new(VulkanContext::new(vulkano_context)),
            vulkano_windows: Arc::new(VulkanoWindows::default()),
            gui: HashMap::new(),
            input_manager: Arc::new(input_manager),
            scene_manager: HashMap::new(),
            timer: Arc::new(Timer::new()),
            event_loop_proxy,
        }
    }
}

impl ApplicationHandler<UserEvent> for App {
    fn resumed(&mut self, event_loop: &ActiveEventLoop) {
        if let Some(vulkano_windows) = Arc::get_mut(&mut self.vulkano_windows) {
            let window_id = vulkano_windows.create_window(
                event_loop,
                self.vulkan_context.vulkano_context(),
                &WindowDescriptor {
                    title: "Rust ASH Test".to_string(),
                    width: 800.0,
                    height: 600.0,
                    present_mode: PresentMode::Fifo,
                    cursor_visible: false,
                    cursor_locked: true,
                    ..Default::default()
                },
                |_| {},
            );

            let renderer = vulkano_windows.get_renderer_mut(window_id).unwrap();
            renderer.add_additional_image_view(
                DEPTH_IMAGE_ID,
                Format::D16_UNORM,
                ImageUsage::DEPTH_STENCIL_ATTACHMENT,
            );

            let gui = {
                Gui::new(
                    event_loop,
                    renderer.surface(),
                    renderer.graphics_queue(),
                    renderer.swapchain_format(),
                    GuiConfig {
                        is_overlay: true,
                        allow_srgb_render_target: true,
                        ..Default::default()
                    },
                )
            };
            self.gui.insert(window_id, gui);

            let mut scene_manager = SceneManager::new();
            scene_manager.load_scene(Box::new(TestScene::default()));

            self.scene_manager.insert(window_id, scene_manager);
        }
    }

    fn device_event(
        &mut self,
        _event_loop: &ActiveEventLoop,
        _device_id: winit::event::DeviceId,
        event: winit::event::DeviceEvent,
    ) {
        match Arc::get_mut(&mut self.input_manager) {
            Some(input_manager) => input_manager.process_device_event(&event),
            None => log::error!("Failed to get a mutable reference to the input manager"),
        }
    }

    fn window_event(&mut self, event_loop: &ActiveEventLoop, id: WindowId, event: WindowEvent) {
        let renderer = Arc::get_mut(&mut self.vulkano_windows)
            .unwrap()
            .get_renderer_mut(id)
            .unwrap();
        let gui = self.gui.get_mut(&id).unwrap();

        if !gui.update(&event) {
            match Arc::get_mut(&mut self.input_manager) {
                Some(input_manager) => input_manager.process_window_event(&event),
                None => log::error!("Failed to get a mutable reference to the input manager"),
            }
        }

        match event {
            WindowEvent::CloseRequested => {
                log::debug!("The close button was pressed; stopping");
                event_loop.exit();
            }
            WindowEvent::Resized(_) => {
                renderer.resize();
            }
            WindowEvent::ScaleFactorChanged { .. } => {
                renderer.resize();
            }
            WindowEvent::RedrawRequested => {
                match Arc::get_mut(&mut self.input_manager) {
                    Some(input_manager) => input_manager.update(),
                    None => log::error!("Failed to get a mutable reference to the input manager"),
                }
                match Arc::get_mut(&mut self.timer) {
                    Some(timer) => timer.update(),
                    None => log::error!("Failed to get a mutable reference to the timer"),
                }

                let mut scene_context = SceneContext::from((
                    &mut *renderer,
                    &self.vulkan_context,
                    &self.input_manager,
                    &self.timer,
                    id,
                    &self.event_loop_proxy,
                ));

                let scene_manager = self.scene_manager.get_mut(&id).unwrap();
                scene_manager
                    .load_scene_if_not_loaded(&scene_context)
                    .unwrap();

                if let Some(scene) = scene_manager.current_scene_mut() {
                    scene.update(&scene_context).unwrap();

                    let acquire_future = renderer.acquire(None, |_| {}).unwrap();
                    // Update the swapchain image view to the current one after acquiring the next image
                    scene_context.swapchain_image_view = renderer.swapchain_image_view();

                    let acquire_future = scene.render(acquire_future, &scene_context, gui).unwrap();
                    renderer.present(acquire_future, true);
                }
            }
            _ => {}
        }
    }

    fn about_to_wait(&mut self, _event_loop: &ActiveEventLoop) {
        let window = self.vulkano_windows.get_primary_window().unwrap();
        window.request_redraw();
    }

    fn user_event(&mut self, event_loop: &ActiveEventLoop, event: UserEvent) {
        match event {
            UserEvent::CursorGrabMode(window_id, grab) => {
                let window = self.vulkano_windows.get_window(window_id).unwrap();
                if let Err(e) = window.set_cursor_grab(grab) {
                    log::error!("Failed to set cursor grab: {}", e);
                }
            }
            UserEvent::CursorVisible(window_id, visible) => {
                let window = self.vulkano_windows.get_window(window_id).unwrap();
                window.set_cursor_visible(visible);
            }
            UserEvent::ChangeScene(window_id, scene) => {
                let scene_manager = self.scene_manager.get_mut(&window_id).unwrap();
                scene_manager.load_scene(scene);
            }
        }
    }
}
